package com.fundi.framework.common.helpers;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;

import com.fundi.framework.common.displaylist.DisplayList;
import com.fundi.framework.common.displaylist.ListFieldDefn;
import com.fundi.framework.common.displaylist.ListFilter;
import com.fundi.framework.common.displaylist.ListFilterCondition;
import com.fundi.framework.common.instance.AppInstance;
import com.fundi.framework.common.nl1.Messages;
import com.fundi.framework.common.util.StringEditor;

public class ListFilterHelper {

	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	private AppInstance instance;
	
	private String listFilterFilenamePfx;
	
	private ArrayList<ListFilter> filters = new ArrayList<ListFilter>();
	
	private DisplayList list = null;
	private String listContext = "";
	
	
	public ListFilterHelper(AppInstance instance, String listContext) {

		super();
		
		this.instance = instance;
		this.listContext = listContext;
		
		setListFilterFilenamePfx(listContext);
	}

	public ListFilterHelper(AppInstance instance, DisplayList list) {

		super();
		
		this.instance = instance;
		this.list = list;
		this.listContext = list.getContext();
		
		setListFilterFilenamePfx(list.getContext());
	}
	
	public void setDisplayList(DisplayList newList) {
		
		ArrayList<ListFilter> previousFilters = new ArrayList<ListFilter>();
		
		//The list is theoretically new at this point, so it won't have any filters.  If it has the same context as the prior list, copy the
		//filters over so they will re-apply to new (refreshed) list.
		
		if (list != null) {
			if (newList.getContext().equals(list.getContext()) && list.getCurrentFilters().size() > 0) {
				previousFilters.addAll(list.getCurrentFilters());
			}
		}
			
		list = newList;
		list.setCurrentFilters(previousFilters);
		setListFilterFilenamePfx(list.getContext());	//this needs to change or the next call to retrieveSavedFilters will retrieve them for the old context
	}
	
	public DisplayList getDisplayList() {
		return list;
	}

	public ArrayList<ListFilter> retrieveSavedFilters() {
		
		ArrayList<File> files = instance.getUserFiles(listFilterFilenamePfx);
		JAXBContext context;
		Unmarshaller u;
		Iterator<File> itFile;
		ListFilter filter;
		int idx;
		ArrayList<ListFilter> listFilters = new ArrayList<ListFilter>();
		
		filters.clear();
		
		if (list != null) {
			listFilters = list.getCurrentFilters();
		}
		
		try {
			itFile = files.iterator();
			while (itFile.hasNext()) {
				context = JAXBContext.newInstance(ListFilter.class);
				u = context.createUnmarshaller();
				filter = (ListFilter) u.unmarshal(new FileInputStream(itFile.next()));
				//If the list already has filters for the same context, retain the active status from the list (since FUD-201).
				if (list != null) {
					if (listContext.equals(list.getContext())) {
						for (idx = 0; idx < listFilters.size(); idx++) {
							if (listFilters.get(idx).getDescription().equals(filter.getDescription())) {
								filter.setActive(listFilters.get(idx).isActive());
								break;
							}
						}
					}
					else {
						listContext = list.getContext();
					}
				}
				filters.add(filter);
			}
		} 
		catch (Throwable e) {
			instance.addConsoleMessage(e);
		}
		Collections.sort(filters);
		
		return filters;
	}
	
	public String saveFilter(ListFilter filter) {
		
		File filtFile;
		JAXBContext context;
		Marshaller m;
		FileOutputStream out;
		ArrayList<ListFilter> listFilters;
		int idx;
		
		filtFile = new File(instance.getWorkingFolder(), listFilterFilenamePfx + filter.getDescription() + Messages.getString("ListFilterHelper_6")); //$NON-NLS-1$
		
		//If the filter is in the list, update it's status in the list as it may have changed.
		if (list != null) {
			listFilters = list.getCurrentFilters();
			if (listContext.equals(list.getContext())) {
				for (idx = 0; idx < listFilters.size(); idx++) {
					if (listFilters.get(idx).getDescription().equals(filter.getDescription())) {
						listFilters.get(idx).setActive(filter.isActive());
					}
				}
			}
			else {
				listContext = list.getContext();
			}
		}
		
		try {
			context = JAXBContext.newInstance(ListFilter.class);
			m = context.createMarshaller();
			m.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
			out = new FileOutputStream(filtFile.getAbsolutePath());
			m.marshal(filter, out);
			out.close();
			return ""; //$NON-NLS-1$
		}
		catch (Throwable e) {
			instance.addConsoleMessage(e);
			return instance.exceptionToString(e);
		}
	}
	
	public String deleteFilter(ListFilter filter) {
		
		File filtFile;
		String msg = ""; //$NON-NLS-1$
		boolean deleted;
		
		filtFile = new File(instance.getWorkingFolder(), listFilterFilenamePfx + filter.getDescription() + Messages.getString("ListFilterHelper_9")); //$NON-NLS-1$
		deleted = filtFile.delete();
		if (!deleted) {
			msg = Messages.getString("ListFilterHelper_10"); //$NON-NLS-1$
		}
		
		return msg;
	}
	
	public String createFilter(String description) {
		
		String msg = ""; //$NON-NLS-1$
		String desc = description.trim();
		ListFilter filter;
		Iterator<ListFilter> itFilter;
		
		if (filters.size() == 0) {
			retrieveSavedFilters();  //Just to be sure - the hosting dialog should have already done this.
		}
		
		if (desc.length() == 0) {
			msg = Messages.getString("ListFilterHelper_12"); //$NON-NLS-1$
		}
		else {
			if (!(new StringEditor(desc).isValidFilename())) {
				msg = Messages.getString("ListFilterHelper_1") + StringEditor.invalidFilenameCharacters; //$NON-NLS-1$
			}
			else {
				for (itFilter = filters.iterator(); itFilter.hasNext();) {
					if (itFilter.next().getDescription().equalsIgnoreCase(desc)) {
						msg = Messages.getString("ListFilterHelper_13"); //$NON-NLS-1$
					}
				}
			}
		}
			
		if (msg.length() == 0) {
			filter = new ListFilter(desc);
			msg = saveFilter(filter);
			retrieveSavedFilters();
		}
		
		return msg;
	}
	
	public String editCondition(DisplayList list, ListFilterCondition originalCondition, String field, String operator, String enteredValue) {
		
		/*
		 *	A filter can be applied to any list (within in the supplied context), and only conditions for fields within that list will
		 *	be used in filtering.
		 *
		 *	If the field can be found in the supplied list, edit against its definition.
		 *	If the field cannot be found in the supplied list, edit against the supplied definition, which should be the one used
		 *	when that condition was last edited against the filter with a list containing the field. 
		 */
		
		String msg = ""; //$NON-NLS-1$
		String value = enteredValue.trim().toUpperCase();
		ListFieldDefn defn = null;
		int idx;
		boolean operatorOK;
		
		StringTokenizer parser = null;
		
		if (field.equals("")) { //$NON-NLS-1$
			msg = Messages.getString("ListFilterHelper_16"); //$NON-NLS-1$
		}
		else {
			if (operator.equals("")) { //$NON-NLS-1$
				msg = Messages.getString("ListFilterHelper_18"); //$NON-NLS-1$
			}
			else {
				if (value.equals("") && (operator.equals(ListFilterCondition.opIn) || operator.equals(ListFilterCondition.opBetween))) { //$NON-NLS-1$
					msg = Messages.getString("ListFilterHelper_20"); //$NON-NLS-1$
				}
			}
		}
		
		if (field.equals(ListFilterCondition.anyField) && !operator.equals(ListFilterCondition.opContains)) { //$NON-NLS-1$
			msg = Messages.getString("ListFilterHelper_29"); //$NON-NLS-1$
		}
		
		if (msg.length() == 0) {
			operatorOK = false;
			for (idx = 0; idx < ListFilterCondition.allowedOperands.length; idx++) {
				if (ListFilterCondition.allowedOperands[idx].equals(operator)) {
					operatorOK = true;
					break;
				}
			}
			if (!operatorOK) {
				msg = Messages.getString("ListFilterHelper_28"); //$NON-NLS-1$
			}
		}
		
		if (msg.length() == 0 && !(field.equals(ListFilterCondition.anyField))) {
			//Find the definition for the selected field.
			defn = getFieldDefinition(list, field);
			if (defn == null) {
				defn = originalCondition.getFieldDefn();  //Use the one the condition was last built/edited with.
			}
			
			if (operator.equals(ListFilterCondition.opIn) || operator.equals(ListFilterCondition.opNotIn)) {
				//Needs to be a comma delimited list.
				parser = new StringTokenizer(value, Messages.getString("ListFilterHelper_21")); //$NON-NLS-1$
				if (parser.countTokens() == 0) {
					msg = Messages.getString("ListFilterHelper_22"); //$NON-NLS-1$
				}
				else {
					if (!(defn.getDataType().equals(ListFieldDefn.typeCharacter))) {
						try {
							while (parser.hasMoreTokens()) {
								if (!okNumber(defn, parser.nextToken())) {
									msg = Messages.getString("ListFilterHelper_23"); //$NON-NLS-1$
									break;
								}
							}
						}
						catch (NoSuchElementException nsee) {}
					}
				}
			}
			else {
				if (operator.equals(ListFilterCondition.opBetween)) {
					//Needs to be a comma delimited list.
					parser = new StringTokenizer(value, Messages.getString("ListFilterHelper_24")); //$NON-NLS-1$
					if (parser.countTokens() != 2) {
						msg = Messages.getString("ListFilterHelper_25"); //$NON-NLS-1$
					}
					else {
						if (!(defn.getDataType().equals(ListFieldDefn.typeCharacter))) {
							try {
								while (parser.hasMoreTokens()) {
									if (!okNumber(defn, parser.nextToken())) {
										msg = Messages.getString("ListFilterHelper_26"); //$NON-NLS-1$
										break;
									}	
								}
							}
							catch (NoSuchElementException nsee) {}
						}
					}
				}
				else {
					if (!(defn.getDataType().equals(ListFieldDefn.typeCharacter))) {
						if (!okNumber(defn, value)) {
							msg = Messages.getString("ListFilterHelper_27"); //$NON-NLS-1$
						}
					}	
				}
			}
		}
		
		return msg;
	}

	private void setListFilterFilenamePfx(String listContext) {
		listFilterFilenamePfx = Messages.getString("ListFilterHelper_4") + listContext + Messages.getString("ListFilterHelper_5"); //$NON-NLS-1$ //$NON-NLS-2$
	}
	
	private boolean okNumber(ListFieldDefn defn, String value) {

		boolean isOK = true;
		
		try {
			if (defn.getDataType().equals(ListFieldDefn.typeNumeric)) {
				Double.parseDouble(value);
				if (value.toLowerCase().contains("d") || value.toLowerCase().contains("f")) {
					isOK = false;  //Java representation of double or float shouldn't be accepted, e.g. 4f or 4d.
				}
			}
		}
		catch (NumberFormatException nfe) {
			isOK = false;
		}
		
		return isOK;
	}
	
	private ListFieldDefn getFieldDefinition(DisplayList list, String label) {
		
		int idxDefn;
		ListFieldDefn defn;
		
		for (idxDefn = 0; idxDefn < list.getFieldDefns().size(); idxDefn++) {
			defn = list.getFieldDefns().get(idxDefn);
			if (defn.getLabel().equalsIgnoreCase(label)) {
				return defn;
			}
		}
		
		return null;
	}
	
	public void clearConditions(ListFilter filter) {
		
		filter.clearConditions();
	}
	
	public void addCondition(ListFilter filter, DisplayList list, ListFilterCondition originalCondition, String field, String operator, String enteredValue,
												boolean highlight, int red, int green, int blue, boolean caseSensitive) {
		
		ListFieldDefn defn = getFieldDefinition(list, field);
		
		if (defn == null) {
			//If the condition is for all fields, there isn't a definition so create a dummy one.
			if (field.equals(ListFilterCondition.anyField)) {
				defn = new ListFieldDefn(ListFilterCondition.anyField, ListFieldDefn.typeCharacter, ListFieldDefn.alignLeft, false, true);
			}
			else {
				defn = originalCondition.getFieldDefn();
			}
		}
		
		filter.addCondition(defn, operator, enteredValue, highlight, red, green, blue, caseSensitive);
	}
	
	public String saveFilter(ListFilter filter, ArrayList<ListFilterCondition> conditions) {
		
		String msg = ""; //$NON-NLS-1$
		
		return msg;
	}
}
